;;;-*- Mode: Lisp; Syntax: Common-Lisp; Base: 10; Package: eval-basic -*-
#|
-----------------------------------------------------------------------------------
TITLE: Syntax definitions and other basic definitions for eval
-----------------------------------------------------------------------------------
File:    eval-basic.em
Version: 1.2 (last modification on Wed Jan 26 09:24:13 1994)
State:   published

DESCRIPTION:

DOCUMENTATION:

NOTES:

REQUIRES:

PROBLEMS:

AUTHOR:
Ingo Mohr

CONTACT: 
Ingo Mohr

HISTORY: 
Log for /export/home/saturn/ukriegel/Dist/Apply/eval-basic.em[1.2]:
  basic stuff for eval (syntax, variable handling)
[1.1] Thu Jan 13 15:49:21 1994 imohr@isst saved
  [Wed Jan 12 16:51:45 1994] Intention for change:
  global variables
  improved error signalling
[1.2] Wed Jan 26 09:25:33 1994 wheick@isst published
  [Fri Jan 14 08:13:35 1994] Intention for change:
  --- no intent expressed ---

-----------------------------------------------------------------------------------
|#

#module eval-basic

(import (eulisp1 
         lzs
         el2lzs-error
         (only (find-module) el2lzs-main)
         (only (mapcar mapc assoc rest first append char-code) 
           common-lisp))

 syntax (eulisp1
         (only (push) common-lisp))

 export (*global-variables*
         initialize-variables)
 )

;;; -----------------------------------------------------------------------------------
;;; syntax for declaring system functions
;;; -----------------------------------------------------------------------------------

(defmacro declare-system-functions functions
  `(list ,@(mapcar (lambda (fun)
                     `(cons ',(make-eulisp-symbol fun)
                            #',fun))
                   functions)))

;;; -----------------------------------------------------------------------------------
;;; basics for the variable environment
;;; -----------------------------------------------------------------------------------

(defvar variable-environment () ) 
;;; holds the current bindings of local lexical variables in an a-list of the form
;;; ((a-lzs-local-static . value) ...).

(deflocal *global-variables* () )
;;; holds the current values of global lexical variables after the value was
;;; changed by setq; until the value is changed the current value must be taken
;;; from the annotation initial-value
;;; the values are stored in an a-list ((a-lzs-global-static . value) ...)

;;; The forms in body are executed with variables in vars bound to the
;;; corresponding values of the LZS-forms in values (computed by eval) and the
;;; variable in rst (unless it is () ) bound to the list of values not used for
;;; the other variables. The macro can be used for handling function
;;; applications and let*-forms of the LZS.

(defmacro with-new-values (vars rst values . body)
  `(dynamic-let ((variable-environment (dynamic variable-environment)))
                (initialize-variables ,vars ,rst ,values)
                ,@body))

(defun initialize-variables (vars rst values)
(cond ((null vars) (initialize-rest rst values))
      ((null values) (eval-error-too-few-arguments))
      (t (push (cons (first vars) (first values))
               (dynamic variable-environment))
         (initialize-variables (rest vars) rst (rest values)))))

(defun initialize-rest (rst values)
  (cond ((and (null rst) values)
         (eval-error-too-many-arguments))
        ((null rst))
        (t (push (cons rst values)
                 (dynamic variable-environment)))))

#module-end
